#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jan 11 14:17:50 2022

@author: m102324
"""

import sys
import logging
import pandas as pd
from cobindability.BED import bed_overlap_size, bedtolist, bedinfo
from cobindability.coefcal import ov_coef, ov_jaccard, ov_ss, ov_sd, pmi_value, npmi_value
from cobindability import version


__author__ = "Liguo Wang"
__copyright__ = "Copyleft"
__credits__ = []
__license__ = "MIT"
__version__ = version.version
__maintainer__ = "Liguo Wang"
__email__ = "wang.liguo@mayo.edu"
__status__ = "Development"


def ov_stats(file1, file2, bg_size = 1400000000):
	"""

	Parameters
	----------
	file1 : str
		Genomic regions in BED (Browser Extensible Data, https://genome.ucsc.edu/FAQ/FAQformat.html#format1), BED-like or BigBed format.
		The BED-like format includes 'bed3','bed4','bed6','bed12','bedgraph','narrowpeak', 'broadpeak','gappedpeak'. BED and BED-like
		format can be plain text, compressed (.gz, .z, .bz, .bz2, .bzip2) or remote (http://, https://, ftp://) files. Do not compress
		BigBed foramt. BigBed file can also be a remote file.
	file2 : str
		Genomic regions in BED (Browser Extensible Data, https://genome.ucsc.edu/FAQ/FAQformat.html#format1), BED-like or BigBed format.
		The BED-like format includes 'bed3','bed4','bed6','bed12','bedgraph','narrowpeak', 'broadpeak','gappedpeak'. BED and BED-like
		format can be plain text, compressed (.gz, .z, .bz, .bz2, .bzip2) or remote (http://, https://, ftp://) files. Do not compress
		BigBed foramt. BigBed file can also be a remote file.
	draws_n : int
		Times of drawing samples with replacement. Set to '0' to turn off bootstraping.
	draw_frac : float, optional
		The size of subsample. The default is 0.85 (85% of the orignal genomic reginos will be selected).
	bg_size : int, optional
		The effective background genome size. About 1.4Gb of the human genome are. The default is 1424655930.

	Returns
	-------
	pandas.Series
		Pandas Series of 'coef_obs', 'coef_exp','coef_ratio', 'coef_ratio_low', 'coef_ratio_high'.
		'coef_obs' : Observed overlap coefficient between two BED files.
		'coef_exp' : Expected overlap coefficient between two BED files.
		'coef_ratio' : Ratio between 'coef_obs' and 'coef_exp'.
		'coef_ratio_low' : The lower bound of 95% confidence interval of 'coef_ratio'.
		'coef_ratio_high' : The upper bound of 95% confidence interval of 'coef_ratio'.


	"""
	results = {}
	file1_lst = bedtolist (file1)
	file2_lst = bedtolist (file2)


	logging.info("Gathering information for \"%s\" ..." % file1)
	info1 = bedinfo(file1)
	results['A.name'] = info1['Name']
	results['A.interval_count'] = info1['Count']
	results['A.interval_total_size'] = info1['Total_size']
	#results['A.interval_uniq_size'] = info1['Genomic_size']
	results['A.interval_mean_size'] = info1['Mean_size']
	results['A.interval_median_size'] = info1['Median_size']
	results['A.interval_min_size'] = info1['Min_size']
	results['A.interval_max_size'] = info1['Max_size']
	results['A.interval_size_SD'] = info1['STD']
	uniqBase1 = info1['Genomic_size']

	logging.info("Gathering information for \"%s\" ..." % file2)
	info2 = bedinfo(file2)
	results['B.name'] = info2['Name']
	results['B.interval_count'] = info2['Count']
	results['B.interval_total_size'] = info2['Total_size']
	#results['B.interval_uniq_size'] = info2['Genomic_size']
	results['B.interval_mean_size'] = info2['Mean_size']
	results['B.interval_median_size'] = info2['Median_size']
	results['B.interval_min_size'] = info2['Min_size']
	results['B.interval_max_size'] = info2['Max_size']
	results['B.interval_size_SD'] = info2['STD']
	uniqBase2 = info2['Genomic_size']

	#calculate overall overlap coef
	logging.debug("Calculating overlapped bases ...")
	overlapBases = bed_overlap_size(file1_lst, file2_lst)

	results['G.size'] = bg_size
	results['A.size'] = uniqBase1
	results['Not_A.size'] = bg_size - uniqBase1
	results['B.size'] = uniqBase2
	results['Not_B.size'] = bg_size - uniqBase2
	results['A_not_B.size'] = uniqBase1 - overlapBases
	results['B_not_A.size'] = uniqBase2 - overlapBases
	results['A_and_B.size'] = overlapBases
	results['A_and_B.exp_size'] = uniqBase1 * uniqBase2/bg_size
	results['A_or_B.size'] = uniqBase1 + uniqBase2 -overlapBases
	results['Neither_A_nor_B.size'] = bg_size - uniqBase1 - uniqBase2 + overlapBases


	results['coef.Overlap'] = ov_coef(uniqBase1, uniqBase2, overlapBases, bg_size)
	results['coef.Jaccard'] = ov_jaccard(uniqBase1, uniqBase2, overlapBases, bg_size)
	results['coef.Dice'] = ov_sd(uniqBase1, uniqBase2, overlapBases, bg_size)
	results['coef.SS'] = ov_ss(uniqBase1, uniqBase2, overlapBases, bg_size)
	results['A_and_B.PMI'] = pmi_value(uniqBase1, uniqBase2, overlapBases, bg_size)
	results['A_and_B.NPMI'] = npmi_value(uniqBase1, uniqBase2, overlapBases, bg_size)


	return pd.Series(data=results)





if __name__ == '__main__':
	pd.set_option('display.float_format', lambda x: '%.4f' % x)
	a = ov_stats(sys.argv[1], sys.argv[2])
	print (a)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 12 22:16:41 2022

@author: m102324
"""
import math
import logging
import numpy as np
from cobindability import version


__author__ = "Liguo Wang"
__copyright__ = "Copyleft"
__credits__ = []
__license__ = "MIT"
__version__ = version.version
__maintainer__ = "Liguo Wang"
__email__ = "wang.liguo@mayo.edu"
__status__ = "Development"

def ov_coef(x, y, xy, g):
	"""
	Calculate the overlap coefficient.

	Parameters
	----------
	x : int
		The cardinality of A.
	y : int
		The cardinality of B.
	xy : int
		The cardinality of A AND B.
	g : int
		The cardinality of background.

	Returns
	-------
	float
		Overlap coefficient.
	"""
	if xy > x or xy > y:
		logging.error("Invalid parameters.")
	if min(x,y,xy) < 0:
		logging.error("Invalid parameters.")
	if x == 0 or y == 0 or xy == 0:
		return 0
	else:
		return xy/(math.sqrt(x * y))

def ov_jaccard(x, y, xy, g):
	"""
	Calculate the Jaccard's coefficient (https://en.wikipedia.org/wiki/Jaccard_index).

	Parameters
	----------
	x : int
		The cardinality of A.
	y : int
		The cardinality of B.
	xy : int
		The cardinality of A AND B.
	g : int
		The cardinality of background.

	Returns
	-------
	float
		Overlap coefficient.
	"""
	if xy > x or xy > y:
		logging.error("Invalid parameters.")
	if min(x,y,xy) < 0:
		logging.error("Invalid parameters.")
	if x == 0 or y == 0:
		return 0
	else:
		return xy/(x + y - xy)

def ov_ss(x, y, xy, g):
	"""
	Calculate the Szymkiewicz–Simpson coefficient, (https://en.wikipedia.org/wiki/Overlap_coefficient).

	Parameters
	----------
	x : int
		The cardinality of A.
	y : int
		The cardinality of B.
	xy : int
		The cardinality of A AND B.
	g : int
		The cardinality of background.

	Returns
	-------
	float
		Szymkiewicz–Simpson coefficient.
	"""
	if xy > x or xy > y:
		logging.error("Invalid parameters.")
	if min(x,y,xy) < 0:
		logging.error("Invalid parameters.")
	if x == 0 or y == 0:
		return 0
	else:
		return xy/min(x,y)

def ov_sd(x, y, xy, g):
	"""
	Calculate the Sørensen–Dice coefficient (https://en.wikipedia.org/wiki/Sørensen–Dice_coefficient).
	Sørensen–Dice coefficient also called "Sørensen–Dice index", "Sørensen index" or "Dice's coefficient".

	Parameters
	----------
	x : int
		The cardinality of A.
	y : int
		The cardinality of B.
	xy : int
		The cardinality of A AND B.
	g : int
		The cardinality of background.

	Returns
	-------
	float
		Overlap coefficient.
	"""
	if xy > x or xy > y:
		logging.error("Invalid parameters.")
	if min(x,y,xy) < 0:
		logging.error("Invalid parameters.")
	if x == 0 or y == 0:
		return 0
	else:
		return 2*xy/(x + y)


def pmi_value(x, y, xy, g):
	"""
	Calculate the pointwise mutual information (PMI).

	Parameters
	----------
	x : int
		The cardinality of A.
	y : int
		The cardinality of B.
	xy : int
		The cardinality of A AND B.
	g : int
		The cardinality of background.

	Returns
	-------
	float
		Pointwise mututal information.
	"""
	if g <= 0:
		logging.error("The cardinality of background must be a postive integer.")
	if xy > x or xy > y:
		logging.error("Invalid parameters.")
	if x > g or y > g:
		logging.error("Invalid parameters.")
	if min(x,y,xy) < 0:
		logging.error("Invalid parameters.")
	px = x/g
	py = y/g
	pxy = xy/g
	if pxy == 0:
		return -np.inf
	else:
		return np.log(pxy) - np.log(px) - np.log(py)

def npmi_value(x, y, xy, g):
	"""
	Calculate the normalized pointwise mutual information (NPMI).

	Parameters
	----------
	x : int
		The cardinality of A.
	y : int
		The cardinality of B.
	xy : int
		The cardinality of A AND B.
	g : int
		The cardinality of background.

	Returns
	-------
	float
		Normalized pointwise mutual information.
	"""
	if g <= 0:
		logging.error("The cardinality of background must be a postive integer.")
	if xy > x or xy > y:
		logging.error("Invalid parameters.")
	if x > g or y > g:
		logging.error("Invalid parameters.")
	if min(x,y,xy) < 0:
		logging.error("Invalid parameters.")
	px = x/g
	py = y/g
	pxy = xy/g
	if pxy == 0:
		return -1
	else:
		return np.log(px*py)/np.log(pxy) - 1


#!/usr/bin/env python
# encoding: utf-8

"""
read bigBed files.

"""


import pyBigWig
from cobindability import version

__author__ = "Liguo Wang"
__copyright__ = "Copyleft"
__credits__ = []
__license__ = "MIT"
__version__ = version.version
__maintainer__ = "Liguo Wang"
__email__ = "wang.liguo@mayo.edu"
__status__ = "Development"


def bbreader(fname):
	"""
	read bigBed file
	"""
	bb = pyBigWig.open(fname)
	chrom_dict = bb.chroms()
	for chr in chrom_dict:
		for start, end, score in bb.entries(chr, 0, chrom_dict[chr]):
			yield(chr + '\t' + str(start) + '\t' + str(end) + '\t' + score)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Dec 11 20:05:15 2021

@author: m102324
"""
import sys
import math
from cobindability.BED import bedtolist,unionBed3,subtractBed3
import logging
import numpy as np
import pandas as pd
from random import sample




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Dec 12 18:50:00 2021

@author: m102324
"""
import os
import logging
import pyBigWig
import pandas as pd
import numpy as np
from scipy.stats import pearsonr, spearmanr, kendalltau
from cobindability import version

__author__ = "Liguo Wang"
__copyright__ = "Copyleft"
__credits__ = []
__license__ = "MIT"
__version__ = version.version
__maintainer__ = "Liguo Wang"
__email__ = "wang.liguo@mayo.edu"
__status__ = "Development"


def bigwig_corr(bed, bw1, bw2, outfile, na_label = 'nan', score_type='mean', exact_scores = True, keep_NA = False, top_x = 1.0, min_sig = 0):
	"""
	Calculate Pearson's and Spearman's correlations between two bigWig files

	Parameters
	----------
	bed : list
		List of genomic regions.
	bw1 : str
		Name of one bigWig file.
	bw2 : str
		Name of another bigWig file.
	outfile : str
		Name of the output TSV file.
	na_label : str, optional
		String representation of missing values. Default: 'nan'
	score_type : str, optional
		Summary statistic score type ('min','mean' or 'max') of a genomic region.
		Default: 'mean'
	exact_scores : bool, optional
		If set, calculate the "exact" summary statistic score rather than
		"zoom-level" score for each genomic region. Default: True
	keep_NA : bool, optional
		If set, all genomic regions will be kept even they do not have summary
		statistic scores in either of the two bigWig files. Default: False
	top_x : float, optional
		Percentage ( if top_x in (0,1]) or number (if top_x > 1) of genomic
		regions used to calculate Pearson, Spearman, Kendall's correlations.
		default: 1.0 (i.e., use all genomic regions to calculate correlation coefficients)
	min_sig : float, optional
		Genomic regions with summary statistic equal or less than this value will be
		filtered out.
		default: 0 (i.e., use all genomic regions to calculate correlation coefficients)


	Returns
	-------
	None.

	"""
	bw_1 = pyBigWig.open(bw1)
	logging.debug(str(bw_1.header()))
	bw_2 = pyBigWig.open(bw2)
	logging.debug(str(bw_2.header()))

	#check bigwig files
	if bw_1.isBigWig():
		pass
	else:
		logging.error("Not a bigWig file: %s" % bw1)
	if bw_2.isBigWig():
		pass
	else:
		logging.error("Not a bigWig file: %s" % bw2)

	#check chrom IDs in bigwig files
	#all_chroms1 = bw_1.chroms().keys()
	#all_chroms2 = bw_2.chroms().keys()

	names=[]
	scores_1 = []
	scores_2 = []
	for (chrom ,start, end) in bed:
		region_id = chrom + ':' + str(start) + '-' + str(end)
		score_1 = bw_1.stats(chrom, start, end, type=score_type, exact=exact_scores).pop()
		score_2 = bw_2.stats(chrom, start, end, type=score_type, exact=exact_scores).pop()
		if (isinstance(score_1, (int, float)) is False):
			if keep_NA:
				score_1 = np.nan
			else:
				continue
		if (isinstance(score_2, (int, float)) is False):
			if keep_NA:
				score_2 = np.nan
			else:
				continue
		names.append(region_id)
		scores_1.append(score_1)
		scores_2.append(score_2)

	bw1_name = os.path.basename(bw1) + '.' + score_type
	bw2_name = os.path.basename(bw2) + '.' + score_type
	df = pd.DataFrame(data={bw1_name : scores_1, bw2_name : scores_2 }, index=names, dtype=float)

	logging.info("Sort dataframe by summary statistical scores ...")
	df.sort_values(by=[bw1_name, bw1_name], ascending=False, ignore_index=False, inplace=True)

	logging.info("Save dataframe to: \"%s\"" % outfile)
	df.to_csv(outfile, na_rep=na_label, sep="\t", index=True, header=True, index_label="region_id")

	#Still remove NAs, in order to calculate correlations
	#if keep_NA:
	df = df[df > min_sig]
	df.dropna(axis=0, inplace=True)

	if top_x > 0 and top_x <= 1:
		top_n = int(len(df)* top_x)
		logging.info("Select %d regions ..." % top_n)
		df = df.head(top_n)
	elif top_x > 1:
		top_n = top_x
		logging.info("Select %d regions ..." % top_n)
		df = df.head(top_n)


	(pearson_cor, pearson_p) = pearsonr(np.log2(df[bw1_name]), np.log2(df[bw2_name]))
	(spearman_rho, spearman_p) = spearmanr(np.log2(df[bw1_name]), np.log2(df[bw2_name]))
	(kendall_tau, kendall_p) = kendalltau(np.log2(df[bw1_name]), np.log2(df[bw2_name]))

	return (pd.DataFrame(data={'Pearson_cor:' : [pearson_cor, pearson_p], 'Spearman_rho:' : [spearman_rho, spearman_p], 'Kendall_tau:' : [kendall_tau, kendall_p]}, index=['Correlation', 'P-value']))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Nov  4 11:34:34 2021

@author: m102324
"""
import os

def findBedFiles(root_path = '.', base_name=False, skip_empty=True):
	"""

	Search bed or bed-like files. Supported file formats:
	'bed', 'bed3','bed4','bed6','bed12','bedgraph','bgr','narrowpeak',
	'broadpeak','gappedpeak', 'bb',' bigbed'

	Parameters
	----------
	root_path : str, optional
		Name of directory to seaerch. The default is the current directory.
	base_name : bool, optional
		Return only the base names of BED files. The default is False.
	skip_empty : bool, optional
		Skip empty BED files. The default is True.

	Returns
	-------
	bed_files : list
		A list of bed or bed-like files.

	"""
	suffix_lib = ['bed', 'bed3','bed4','bed6','bed12','bedgraph','bgr','narrowpeak', 'broadpeak','gappedpeak', 'bb',' bigbed']
	bed_files = []
	for root, dirs, files in os.walk(root_path,followlinks=True):
		for f in files:
			suf = f.split('.')[-1]
			if suf.lower() in suffix_lib:
				long_name = os.path.join(root, f)
				if skip_empty and (os.stat(long_name).st_size == 0):
					continue
				if base_name:
					bed_files.append(f)
				else:
					bed_files.append(long_name)
	return bed_files


#!/usr/bin/env python

import sys,os
import logging
import numpy as np
import pandas as pd
from os.path import basename
from scipy.stats import fisher_exact
from bx.bitset_builders import binned_bitsets_from_file, binned_bitsets_from_list
from bx.intervals.intersection import Interval, Intersecter
from cobindability import ireader, version

__author__ = "Liguo Wang"
__copyright__ = "Copyleft"
__credits__ = []
__license__ = "MIT"
__version__ = version.version
__maintainer__ = "Liguo Wang"
__email__ = "wang.liguo@mayo.edu"
__status__ = "Development"


def unionBed3(inbed):
	"""
	Merge/union genomic intervals. If input is a BED file, only consider the
	first three columns (chrom, start, end), other columns will be ignored.

	Parameters
	----------
	inbed : str or list
		Name of a BED file or list of genomic intervals (for example,
		[(chr1 100 200), (chr2 150  300), (chr2 1000 1200)])

	Returns
	-------
	unioned_intervals : list
		List of genomic intervals with the overlapped regions merged.

	Examples
	--------
	>>> unionBed3([('chr1', 1, 10), ('chr1', 3, 15), ('chr1', 20, 35), ('chr1', 20, 50)])
	[('chr1', 1, 15), ('chr1', 20, 50)]

	"""
	unioned_intervals = []
	if type(inbed) is list:
		if len(inbed) == 0:
			return unioned_intervals
		else:
			bitsets = binned_bitsets_from_list(inbed)
	elif type(inbed) is str:
		try:
			bitsets = binned_bitsets_from_file( ireader.reader(inbed) )
		except:
			logging.error("invalid input: %s" % inbed)
			sys.exit(1)
	else:
		logging.error("invalid input: %s" % inbed)
		sys.exit(1)

	for chrom in bitsets:
		bits = bitsets[chrom]
		end = 0
		while 1:
			start = bits.next_set( end )
			if start == bits.size: break
			end = bits.next_clear( start )
			unioned_intervals.append((chrom, start, end))
	bitsets=dict()
	return unioned_intervals


def intersectBed3(inbed1,inbed2):
	"""
	Return the shared genomic intervals beetween inbed1 and inbed2. Inputs are
	two BED files or two lists of genomic intervals. If input is a BED file,
	only consider the first three columns (chrom, start, end), other columns
	will be ignored.

	Parameters
	----------
	inbed1 : str or list
		Name of a BED file or list of genomic intervals, for example,
		[(chr1 100 200), (chr2 1000 1200)]
	inbed2 : str or list
		Name of a BED file or list of genomic intervalss, for example,
		[(chr1 150 220), (chr2 1100 1300)]

	Returns
	-------
	shared_intervals : list
		List of genomic intervals shared between the two input BED files (or lists).

	Examples
	--------
	>>> intersectBed3([('chr1', 1, 10), ('chr1', 20, 35)], [('chr1',3, 15), ('chr1',20, 50)])
	[('chr1', 3, 10), ('chr1', 20, 35)]
	"""
	shared_intervals = []
	# read inbed1
	if type(inbed1) is list:
		if len(inbed1) == 0:
			return shared_intervals
		else:
			bits1 = binned_bitsets_from_list(inbed1)
	elif type(inbed1) is str:
		try:
			bits1 = binned_bitsets_from_file(ireader.reader(inbed1))
		except:
			logging.error("invalid input: %s" % inbed1)
			sys.exit(1)
	else:
		logging.error("invalid input: %s" % inbed1)
		sys.exit(1)

	# read inbed2
	if  type(inbed2) is list:
		if len(inbed2) == 0:
			return shared_intervals
		else:
			bits2 = binned_bitsets_from_list(inbed2)
	elif  type(inbed2) is str:
		try:
			bits2 = binned_bitsets_from_file(ireader.reader(inbed2))
		except:
			logging.error("invalid input: %s" % inbed2)
			sys.exit(1)
	else:
		logging.error("invalid input: %s" % inbed2)
		sys.exit(1)
	bitsets = dict()

	# determine the shared intervals
	for key in bits1:
		if key in bits2:
			bits1[key].iand( bits2[key] )
			bitsets[key] = bits1[key]
	for chrom in bitsets:
		bits = bitsets[chrom]
		end = 0
		while 1:
			start = bits.next_set( end )
			if start == bits.size: break
			end = bits.next_clear( start )
			shared_intervals.append((chrom, start, end))
	bits1.clear()
	bits2.clear()
	bitsets.clear()
	return shared_intervals


def subtractBed3(inbed1,inbed2):
	"""
	Subtract inbed2 from inbed1 (inbed1 - inbed2)

	Parameters
	----------
	inbed1 : str or list
		Name of a BED file or list of genomic intervals, for example,
		[(chr1 100 200), (chr2 1000 1200)]
	inbed2 : str or list
		Name of a BED file or list of genomic intervals, for example,
		[(chr1 150 220), (chr2 1100 1300)]

	Returns
	-------
	remain_intervals : list
		List of genomic intervals from inbed1 with those shared regions with inbed2 removed.

	Examples
	--------
	>>> subtractBed3([('chr1', 1, 10), ('chr1', 20, 35)], [('chr1',3, 15), ('chr1',20, 50)])
	[('chr1', 1, 3)]

	"""
	remain_intervals = []
	# read inbed1
	if type(inbed1) is list:
		if len(inbed1) == 0:
			return remain_intervals
		else:
			bitsets1 = binned_bitsets_from_list(inbed1)
	elif type(inbed1) is str:
		try:
			bitsets1 = binned_bitsets_from_file(ireader.reader(inbed1))
		except:
			logging.error("invalid input: %s" % inbed1)
			sys.exit(1)
	else:
		logging.error("invalid input: %s" % inbed1)
		sys.exit(1)
	# read inbed2
	if type(inbed2) is list:
		if len(inbed2) == 0:
			if type(inbed1) is list:
				return inbed1
			else:
				return bedtolist(inbed1)
		else:
			bitsets2 = binned_bitsets_from_list(inbed2)
	elif  type(inbed2) is str:
		try:
			bitsets2 = binned_bitsets_from_file(ireader.reader(inbed2))
		except:
			logging.error("invalid input: %s" % inbed2)
			sys.exit(1)
	else:
		logging.error("invalid input: %s" % inbed2)
		sys.exit(1)
	for chrom in bitsets1:
		if chrom not in bitsets1:
			continue
		bits1 = bitsets1[chrom]
		if chrom in bitsets2:
			bits2 = bitsets2[chrom]
			bits2.invert()
			bits1.iand( bits2 )
		end=0
		while 1:
			start = bits1.next_set( end )
			if start == bits1.size: break
			end = bits1.next_clear( start )
			remain_intervals.append((chrom,start,end))
	bitsets1 = dict()
	bitsets2 = dict()
	return remain_intervals

def bed_actual_size(*argv):
	'''
	Calculate the aggregated size of genomic intervals.

	Parameters
	----------
	argv : list of genomic regions.
		Each argument can be a list, BED-like file, or a bigBed file. BED file
		can be regular, compressed, or remote file. The suffix of bigBed file
		must be one of ('.bb','.bigbed','.bigBed','.BigBed', '.BB',' BIGBED').

	Returns
	-------
	List of aggregated size.

	Examples
	--------
	>>> intervals_1 = [('chr1', 1, 10), ('chr1', 20, 35)]
	>>> intervals_2 = [('chr1',3, 15), ('chr1',20, 50)]
	>>> bed_actual_size(intervals_1, intervals_2)
	[24, 42]
	'''
	sizes = []
	for arg in argv:
		size = 0
		if type(arg) is list:
			for chrom,start,end in arg:
				size += (int(end) - int(start))
		elif type(arg) is str:
			for l in ireader.reader(arg):
				if l.startswith(('browser','#','track')):continue
				f = l.split()
				if len(f) < 3:
					logging.warning("invalid BED line: %s" % l)
					continue
				tmp = int(f[2]) - int(f[1])
				if tmp <= 0:
					continue
				size += tmp
		sizes.append(size)
	return sizes


def bed_counts(*argv):
	'''
	Calculate the number of genomic intervals in BED file.


	Parameters
	----------
	argv : list of genomic regions.
		Each argument can be a list, BED-like file, or a bigBed file. BED file
		can be regular, compressed, or remote file. The suffix of bigBed file
		must be one of ('.bb','.bigbed','.bigBed','.BigBed', '.BB',' BIGBED').

	Returns
	-------
	List of aggregated size.

	Examples
	--------
	>>> bed1 = [('chr1', 1, 10), ('chr1', 20, 35)]
	>>> bed2 = [('chr1',3, 15), ('chr1',20, 50), ('chr2',100,200)]
	>>> bed_counts(bed1, bed2)
	[2, 3]
	'''
	bed_counts = []
	for arg in argv:
		count = 0
		if type(arg) is list:
			count = len(arg)
		elif type(arg) is str:
			for l in ireader.reader(arg):
				if l.startswith(('browser','#','track')):continue
				f = l.split()
				if len(f) < 3:
					logging.warning("invalid BED line: %s" % l)
					continue
				count += 1
		bed_counts.append(count)
	return bed_counts


def bed_genomic_size(*argv):
	'''
	Calculate the *genomic/unique size* of BED files (or lists of genomic intervals).
	Note, genomic_size <= actual_size.

	Parameters
	----------
	argv : list of genomic regions.
		Each argument can be a list, BED-like file, or a bigBed file. BED file
		can be regular, compressed, or remote file. The suffix of bigBed file
		must be one of ('.bb','.bigbed','.bigBed','.BigBed', '.BB',' BIGBED').

	Returns
	-------
	List of genomic size.

	Example
	-------
	>>> bed1 = [('chr1', 0, 100), ('chr1', 50, 150), ('chr1', 80, 180)]
	>>> bed_genomic_size(bed1)
	[180]
	>>> bed2 = [('chr1', 0, 100), ('chr2', 50, 150), ('chr3', 80, 180)]
	>>> bed_genomic_size(bed2)
	[300]
	>>> bed_genomic_size(bed1, bed2)
	[180, 300]
	'''
	union_sizes = []
	for arg in argv:
		if type(arg) is list:
			bitsets = binned_bitsets_from_list(arg)
		elif type(arg) is str:
			try:
				bitsets = binned_bitsets_from_file( ireader.reader(arg) )
			except:
				logging.error("Invalid input: %s" % arg)
				sys.exit(1)
		else:
			logging.error("Invalid input: %s" % arg)
			sys.exit(1)
		union_size = 0
		for chrom in bitsets:
			bits = bitsets[chrom]
			end = 0
			while 1:
				start = bits.next_set( end )
				if start == bits.size: break
				end = bits.next_clear( start )
				union_size += (end - start)
		union_sizes.append(union_size)
	return (union_sizes)

def bed_overlap_size(bed1,bed2):
	"""
	Calculate the total number of *bases* overlapped between two bed files or
	two lists of genomic intervals.

	Parameters
	----------
	bed1 : str or list
		File name of the first BED file. Can also be a list, such as
		[(chr1 100 200), (chr2 150  300), (chr2 1000 1200)]
	bed2 : str or list
		File name of the second BED file. Can also be a list, such as
		[(chr1 100 200), (chr2 150  300), (chr2 1000 1200)]

	Returns
	-------
	Int. Overlapped size.

	"""
	overlap_size = 0
	if type(bed1) is list:
		bits1 = binned_bitsets_from_list( bed1 )
	else:
		bits1 = binned_bitsets_from_file( ireader.reader(bed1) )
	if type(bed2) is list:
		bits2 = binned_bitsets_from_list( bed2 )
	else:
		bits2 = binned_bitsets_from_file( ireader.reader(bed2) )
	bitsets = dict()
	for key in bits1:
		if key in bits2:
			bits1[key].iand( bits2[key] )
			bitsets[key] = bits1[key]
	for chrom in bitsets:
		bits = bitsets[chrom]
		end = 0
		while 1:
			start = bits.next_set( end )
			if start == bits.size: break
			end = bits.next_clear( start )
			overlap_size += end - start
	return overlap_size

def bedinfo(infile):
	"""
	Basic information of genomic intervals.
	"""
	logging.debug("Gathering teh basic statistics of BED file: %s" % infile)
	bed_infor={}
	bed_infor['Name'] = basename(infile)
	bed_infor['Genomic_size'] = bed_genomic_size(infile)[0]
	bed_infor['Total_size'] = 0
	bed_infor['Count'] = 0
	size = 0
	sizes = []
	for l in ireader.reader(infile):
		if l.startswith(('browser','#','track')):
			continue
		f = l.split()
		if len(f) < 3:
			logging.error("invalid BED line: %s" % l)
		if (int(f[2]) - int(f[1])) < 0:
			logging.error("invalid BED line: %s" % l)
		bed_infor['Count'] += 1
		size = int(f[2]) - int(f[1])
		bed_infor['Total_size'] += size
		sizes.append(size)
	bed_infor['Mean_size'] = np.mean(sizes)
	bed_infor['Median_size'] = np.median(sizes)
	bed_infor['Min_size'] = np.min(sizes)
	bed_infor['Max_size'] = np.max(sizes)
	bed_infor['STD'] = np.std(sizes, ddof=1) #

	return bed_infor


def bedtolist(bedfile):
	"""
	Convert BED file into a list.
	"""
	regions = []
	#print >>sys.stderr, "reading %s ..." %	 bedfile1
	for l in ireader.reader(bedfile):
		l = l.strip()
		if l.startswith('browser') or l.startswith('#') or l.startswith('track'):
			continue
		f = l.split()
		if len(f) < 3:
			logging.error("invalid BED line: %s" % l)
			sys.exit(1)
		if (int(f[2]) - int(f[1])) < 0:
			logging.error("invalid BED line: %s" % l)
			sys.exit(1)
		regions.append(  (f[0], int(f[1]), int(f[2]))  )
	return regions


def compare_bed(inbed1, inbed2):

	"""
	Compare two BED files (or lists). This function is similar to Linux "comm" command.

	Parameters
	----------
	inbed1 : str or list
		Name of a BED file or list of BED regions, for example,
		[(chr1 100 200), (chr2 150  300), (chr2 1000 1200)]
	inbed2 : str or list
		Name of a BED file or list of BED regions, for example,
		[(chr1 100 200), (chr2 150  300), (chr2 1000 1200)]

	Returns
	-------
	bed1_uniq : list
		Genomic regions that are inbed1 unique (i.e., regions only present in inbed1 but
		do not overlap with any regions in inbed2).
	bed2_uniq : list
		Genomic regions that are inbed2 unique (i.e., regions only present in inbed2 but
		do not overlap with any regions in inbed1).
	common : list
		Genomic regions overlapped between inbed1 and inbed2. Note, the
		overlapped regions were merged. For example, (chr1 1 10) and (chr1 5 15)
		will be merged as (chr1 1 15).

	Note
	----
	Overlapped regions *within* input BED files (or lists) are merged before
	comparison.

	"""

	logging.info("Read and union BED file: \"%s\"" % inbed1)
	bed1_union = unionBed3(inbed1)
	#logging.info("Original regions of %s : %d" % (inbed1, len(inbed1)))
	logging.info("Unioned regions of \"%s\" : %d" % (inbed1, len(bed1_union)))

	logging.info("Read and union BED file: \"%s\"" % inbed2)
	bed2_union = unionBed3(inbed2)
	#logging.info("Original regions of %s : %d" % (inbed2, len(inbed2)))
	logging.info("Unioned regions of \"%s\" : %d" % (inbed2, len(bed2_union)))

	logging.info("Merge BED files \"%s\" and \"%s\"" % (inbed1, inbed2))
	bed12_union = unionBed3(bed1_union + bed2_union)
	logging.info("Unioned regions of two BED files : %d" % len(bed12_union))


	logging.info("Build interval tree for unioned BED file: \"%s\"" % inbed1)
	maps1 = {}
	for (ichr1, istart1, iend1) in bed1_union:
		if ichr1 not in maps1:
			maps1[ichr1] = Intersecter()
		maps1[ichr1].add_interval( Interval(istart1, iend1))

	logging.info("Build interval tree for unioned BED file: \"%s\"" % inbed2)
	maps2 = {}
	for (ichr2, istart2, iend2) in bed2_union:
		if ichr2 not in maps2:
			maps2[ichr2] = Intersecter()
		maps2[ichr2].add_interval( Interval(istart2, iend2))

	logging.info("Find common and specific regions ...")
	bed1_uniq = []
	bed2_uniq = []
	common = []
	for (chrom, start, end) in bed12_union:
		if chrom in maps1 and chrom in maps2:
			#found in maps1
			if len( maps1[chrom].find(start, end) ) > 0:
				#found in maps2
				if len( maps2[chrom].find(start, end) ) > 0:
					common.append((chrom, start, end))
				#not found in maps2
				else:
					bed1_uniq.append((chrom, start, end))
			#not found in maps1
			else:
				#found in maps2
				if len( maps2[chrom].find(start, end) ) > 0:
					bed2_uniq.append((chrom, start, end))
				#not found in maps2
				else:
					continue
		elif chrom in maps1:
			bed1_uniq.append((chrom, start, end))
		elif  chrom in maps2:
			bed2_uniq.append((chrom, start, end))
		else:
			continue
	logging.info("\"%s\" unique regions: %d" % (inbed1, len(bed1_uniq)))
	logging.info("\"%s\" unique regions: %d" % (inbed2, len(bed2_uniq)))
	logging.info("Common (overlapped) regions: %d" % len(common))
	return (bed1_uniq, bed2_uniq, common)


def peakwise_ovcoef(inbed1, inbed2, score_func, g, na_label='NA'):

	"""
	Calculates peak-wise overlap .

	Parameters
	----------
	inbed1 : str
		Name of a BED file.
	inbed2 : str
		Name of another BED file.
	score_func : function
		Function to calculate overlap index. Include ov_coef, ov_jaccard, ov_ss, ov_sd.
	g : int
		Size of the genomic background.
	na_label : str
		String label used to represent missing value.

	Returns
	-------
	None
	"""
	#pattern = re.compile(".bed$", re.IGNORECASE)
	logging.info("Read and union BED file: \"%s\"" % inbed1)
	bed1_union = unionBed3(inbed1)
	logging.info("Unioned regions of \"%s\" : %d" % (inbed1, len(bed1_union)))

	logging.info("Read and union BED file: \"%s\"" % inbed2)
	bed2_union = unionBed3(inbed2)
	logging.info("Unioned regions of \"%s\" : %d" % (inbed2, len(bed2_union)))

	#logging.info("Merge BED files \"%s\" and \"%s\"" % (inbed1, inbed2))
	#bed12_union = unionBed3(bed1_union + bed2_union)
	#logging.info("Unioned regions of two BED files : %d" % len(bed12_union))

	logging.info("Build interval tree for unioned BED file: \"%s\"" % inbed1)
	maps1 = {}
	for (ichr1, istart1, iend1) in bed1_union:
		if ichr1 not in maps1:
			maps1[ichr1] = Intersecter()
		maps1[ichr1].add_interval( Interval(istart1, iend1))

	logging.info("Build interval tree for unioned BED file: \"%s\"" % inbed2)
	maps2 = {}
	for (ichr2, istart2, iend2) in bed2_union:
		if ichr2 not in maps2:
			maps2[ichr2] = Intersecter()
		maps2[ichr2].add_interval( Interval(istart2, iend2))

	#overlap bed file 1 with bed file 2
	logging.info("Calculate the overlap coefficient of each genomic region in %s ..." % inbed1)
	outfile_name1 =  os.path.basename(inbed1) + '_peakwise_scores.tsv'
	BED1OUT = open(outfile_name1, 'w')
	print('\t'.join(['chrom','start','end','A.size','B.size', 'A∩B', 'A∪B','B.list', 'Score']), file=BED1OUT)
	for chrom, start, end in bed1_union:
		try:
			bed_1_size = end - start
			if bed_1_size <= 0:
				logging.debug("Skip %s" % (chrom + ':' + str(start) + '-' + str(end)))
			bed_1_lst = [(chrom, start, end)]
			bed_2_size = 0
			bed_2_lst = []

			overlaps = maps2[chrom].find(start, end)
			#print (overlaps)
			if len(overlaps) == 0:
				print('\t'.join([str(i) for i in (chrom, start, end, bed_1_size, na_label, na_label, na_label, na_label, na_label)]), file=BED1OUT)
			else:
				for o in overlaps:
					bed_2_size += (o.end - o.start)
					bed_2_lst.append((chrom, o.start, o.end))
				overlap_size = bed_overlap_size(bed_1_lst, bed_2_lst)
				union_size = bed_genomic_size(bed_1_lst + bed_2_lst)[0]
				peak_ov_coef = score_func(bed_1_size, bed_2_size, overlap_size, g)
				target_list = ','.join([i[0] + ':' + str(i[1]) + '-' + str(i[2]) for i in bed_2_lst])
				print('\t'.join([str(i) for i in (chrom, start, end, bed_1_size, bed_2_size, overlap_size, union_size, target_list, peak_ov_coef)]), file=BED1OUT)
		except:
			print('\t'.join([str(i) for i in (chrom, start, end, bed_1_size, na_label, na_label, na_label, na_label, na_label)]), file=BED1OUT)
	BED1OUT.close()
	logging.info("Save peakwise scores to %s ..." % outfile_name1)

	#overlap bed file 2 with bed file 1
	logging.info("Calculate the overlap coefficient of each genomic region in %s ..." % inbed2)
	outfile_name2 =  os.path.basename(inbed2) + '_peakwise_scores.tsv'
	BED2OUT = open(outfile_name2, 'w')
	print('\t'.join(['chrom','start','end','A.size','B.size', 'A∩B', 'A∪B','B.list', 'Score']), file=BED2OUT)
	for chrom, start, end in bed2_union:
		try:
			bed_2_size = end - start
			if bed_2_size <= 0:
				logging.debug("Skip %s" % (chrom + ':' + str(start) + '-' + str(end)))
			bed_2_lst = [(chrom, start, end)]
			bed_1_size = 0
			bed_1_lst = []

			overlaps = maps1[chrom].find(start, end)
			if len(overlaps) == 0:
				print('\t'.join([str(i) for i in (chrom, start, end, bed_1_size, na_label, na_label, na_label, na_label, na_label)]), file=BED2OUT)
			else:
				for o in overlaps:
					bed_1_size += (o.end - o.start)
					bed_1_lst.append((chrom, o.start, o.end))
				overlap_size = bed_overlap_size(bed_2_lst, bed_1_lst)
				peak_ov_coef = score_func(bed_1_size, bed_2_size, overlap_size, g)
				target_list = ','.join([i[0] + ':' + str(i[1]) + '-' + str(i[2]) for i in bed_1_lst])
				print('\t'.join([str(i) for i in (chrom, start, end, bed_1_size, bed_2_size, overlap_size, union_size, target_list, peak_ov_coef)]), file=BED2OUT)
		except:
			print('\t'.join([str(i) for i in (chrom, start, end, bed_1_size, na_label, na_label, na_label, na_label, na_label)]), file=BED2OUT)
	BED2OUT.close()
	logging.info("Save peakwise scores to %s ..." % outfile_name2)

def cooccur_peak(inbed1, inbed2, inbed_bg, outfile, n_cut=1, p_cut=0.0):

	"""
	Evaluate if two peak sets are significantly oc-occurred or mutually exclusive.
	Using Fisher's exact test.

	Parameters
	----------
	inbed1 : str
		Name of a BED file.
	inbed2 : str
		Name of another BED file.
	inbed_bg : str
		Name of the background BED file (e.g., all promoters, all enhancers).
	outfile : str
		Name of the output file.
	n_cut : int, optional
		Threshold of overlap size. For example, the overlap size is 20 for these two
		regions ('chr1', 0, 100) and ('chr1', 80, 250).
		default = 1
	p_put : float, optional
		Threshold of overlap percentage. In the example above, the overlap percentage
		for ('chr1', 0, 100) is 20/100 = 0.2.
		default = 0.0
	Returns
	-------
	None
	"""
	results = {}
	inbed1_name = os.path.basename(inbed1)
	inbed2_name = os.path.basename(inbed2)
	results['A.name'] = inbed1_name
	results['B.name'] = inbed2_name
	logging.info("Read and union BED file: \"%s\"" % inbed1)
	bed1_union = unionBed3(inbed1)
	results['A.count'] =  len(bed1_union)

	logging.info("Read and union BED file: \"%s\"" % inbed2)
	bed2_union = unionBed3(inbed2)
	results['B.count'] =  len(bed2_union)

	#if inbed_bg is None:
	#	logging.info("Merge two input BED files \"%s\" and \"%s\" as the background" % (inbed1, inbed2))
	#	background = unionBed3(bed1_union + bed2_union)
	#	#logging.info("Number of unioned background regions : %d" % (len(background)))
	#	results['G.count'] =  len(background)
	#else:
	logging.info("Read and union background BED file: \"%s\"" % inbed_bg)
	background = unionBed3(inbed_bg)
	results['G.count'] =  len(background)


	logging.info("Build interval tree for : \"%s\"" % inbed1)
	maps1 = {}
	for (ichr1, istart1, iend1) in bed1_union:
		if ichr1 not in maps1:
			maps1[ichr1] = Intersecter()
		maps1[ichr1].add_interval( Interval(istart1, iend1))

	logging.info("Build interval tree for: \"%s\"" % inbed2)
	maps2 = {}
	for (ichr2, istart2, iend2) in bed2_union:
		if ichr2 not in maps2:
			maps2[ichr2] = Intersecter()
		maps2[ichr2].add_interval( Interval(istart2, iend2))

	#background regions will be divided into 4 categories
	bed1_only = 0
	bed2_only = 0
	cooccur = 0
	neither = 0
	OUT = open(outfile, 'w')


	for chrom, start, end in background:
		line = chrom + '\t' + str(start) + '\t' + str(end)
		bed1_flag = False
		bed2_flag = False

		if (chrom not in maps1) and (chrom not in maps2):
			pass
		elif (chrom not in maps1) and (chrom in maps2):
			bed2_overlaps = maps2[chrom].find(start, end)
			if len(bed2_overlaps) == 0:
				pass
			else:
				bed2_overlap_lst = []
				for o in bed2_overlaps:
					bed2_overlap_lst.append((chrom, o.start, o.end))
				bed2_overlap_size = bed_overlap_size([(chrom, start, end)], bed2_overlap_lst)
				bed2_genomic_size = bed_genomic_size(bed2_overlap_lst)
				try:
					bed2_overlap_ratio = bed2_overlap_size/bed2_genomic_size
				except:
					bed2_overlap_ratio = 0
				if bed2_overlap_size >= n_cut and bed2_overlap_ratio >= p_cut:
					bed2_flag = True
		elif (chrom in maps1) and (chrom not in maps2):
			bed1_overlaps = maps1[chrom].find(start, end)
			if len(bed2_overlaps) == 0:
				pass
			else:
				bed1_overlap_lst = []
				for o in bed1_overlaps:
					bed1_overlap_lst.append((chrom, o.start, o.end))
				bed1_overlap_size = bed_overlap_size([(chrom, start, end)], bed1_overlap_lst)
				bed1_genomic_size = bed_genomic_size(bed1_overlap_lst)
				try:
					bed1_overlap_ratio = bed1_overlap_size/bed1_genomic_size
				except:
					bed1_overlap_ratio = 0
				if bed1_overlap_size >= n_cut and bed1_overlap_ratio >= p_cut:
					bed1_flag = True
		else:
			#overlaps with inbed1
			bed1_overlaps = maps1[chrom].find(start, end)
			#overlaps with inbed2
			bed2_overlaps = maps2[chrom].find(start, end)
			if len(bed1_overlaps) == 0:
				if len(bed2_overlaps) == 0:
					pass
				elif len(bed2_overlaps) > 0:
					bed2_overlap_lst = []
					for o in bed2_overlaps:
						bed2_overlap_lst.append((chrom, o.start, o.end))
					bed2_overlap_size = bed_overlap_size([(chrom, start, end)], bed2_overlap_lst)
					bed2_genomic_size = bed_genomic_size(bed2_overlap_lst)
					try:
						bed2_overlap_ratio = bed2_overlap_size/bed2_genomic_size
					except:
						bed2_overlap_ratio = 0
					if bed2_overlap_size >= n_cut and bed2_overlap_ratio >= p_cut:
						bed2_flag = True
			elif len(bed1_overlaps) > 0:
				bed1_overlap_lst = []
				for o in bed1_overlaps:
					bed1_overlap_lst.append((chrom, o.start, o.end))
				bed1_overlap_size = bed_overlap_size([(chrom, start, end)], bed1_overlap_lst)
				bed1_genomic_size = bed_genomic_size(bed1_overlap_lst)
				try:
					bed1_overlap_ratio = bed1_overlap_size/bed1_genomic_size
				except:
					bed1_overlap_ratio = 0
				if bed1_overlap_size >= n_cut and bed1_overlap_ratio >= p_cut:
					bed1_flag = True

				if len(bed2_overlaps) == 0:
					pass
				elif len(bed2_overlaps) > 0:
					bed2_overlap_lst = []
					for o in bed2_overlaps:
						bed2_overlap_lst.append((chrom, o.start, o.end))
					bed2_overlap_size = bed_overlap_size([(chrom, start, end)], bed2_overlap_lst)
					bed2_genomic_size = bed_genomic_size(bed2_overlap_lst)
					try:
						bed2_overlap_ratio = bed2_overlap_size/bed2_genomic_size
					except:
						bed2_overlap_ratio = 0
					if bed2_overlap_size >= n_cut and bed2_overlap_ratio >= p_cut:
						bed2_flag = True
		if bed1_flag:
			if bed2_flag:
				cooccur += 1
				print (line + '\tCooccur', file=OUT)
			else:
				bed1_only += 1
				print (line + '\t%s_only' % inbed1_name, file=OUT)
		else:
			if bed2_flag:
				bed2_only += 1
				print (line + '\t%s_only' % inbed2_name, file=OUT)
			else:
				neither += 1
				print (line + '\tNeither', file=OUT)
	#print (bed1_only, bed2_only, cooccur, neither)

	results['A+,B-'] = bed1_only
	results['A-,B+'] = bed2_only
	results['A+,B+'] = cooccur
	results['A-,B-'] = neither

	#results['Jaccard index'] = cooccur/(cooccur + neither + bed1_only + bed2_only)
	if bed1_only > bed2_only:
		table = np.array([[neither, bed1_only], [bed2_only, cooccur]])
	else:
		table = np.array([[neither, bed2_only], [bed1_only, cooccur]])
	#print (table)
	oddsr,p = fisher_exact(table, alternative='greater')
	results['odds-ratio'] = oddsr
	results['p-value'] = p
	return pd.Series(data=results, name = "Fisher's exact test result")


def bedtofile(bed_list, bed_file):
	''' Save list of genomic regions to file'''
	OUT = open(bed_file,'w')
	for tmp in bed_list:
		print ('\t'.join([str(i) for i in tmp]), file=OUT)
	OUT.close()


def is_overlap(chr1, st1, end1, chr2, st2, end2):
	'''
	Check if two regios are overlap.

	Parameters
	----------
	chr1 : str
		Chromosome ID of the first genomic region
	st1 : int
		Start coordinate of the first genomic region
	end1 : int
		End coordinate of the first genomic region
	chr2 : str
		Chromosome ID of the second genomic region
	st2 : int
		Start coordinate of the second genomic region
	end2 : int
		End coordinate of the second genomic region
	Return
	------
		0: non overlap
		positive integer [1,): overlap
	'''
	#genome coordinate is left-open, right-close.
	st1 = st1 + 1
	end1 = end1
	st2 = st2 + 1
	end2 = end2
	if chr1 != chr2:
		return 0
	else:
		return len(range(max(st1, st2), min(end1, end2)+1))

def srogcode(lst1, lst2):
	"""
	Determine the spatial relations of genomic regions (SROG)

	Parameters
	----------
	lst1 : tuple
		A tuple of genomic interval. Eg. ('chr1', 1, 100).
	lst2 : tuple
		A tuple of genomic interval. Eg. ('chr1', 15, 60)..

	Returns
	-------
	return_code : str
		Return one of ('disjoint','touch','equal','overlap','contain','within').

	"""
	return_code = ''
	try:
		chrom_1 = lst1[0]
		start_1 = int(lst1[1])
		end_1 = int(lst1[2])
		chrom_2 = lst2[0]
		start_2 = int(lst2[1])
		end_2 = int(lst2[2])
	except:
		return_code = 'unknown'
		return return_code
	ov_size = is_overlap(chrom_1, start_1, end_1, chrom_2, start_2, end_2)
	if ov_size == 0:
		if chrom_1 != chrom_2:
			return_code = 'disjoint'
		else:
			if start_1 == end_2 or end_1 == start_2:
				return_code = 'touch'
			else:
				return_code = 'disjoint'
	elif ov_size > 0:
		if start_1 == start_2 and end_1 == end_2:
			return_code = 'equal'
		elif start_1 >= start_2 and end_1 < end_2:
			return_code = 'within'
		elif start_1 > start_2 and end_1 <= end_2:
			return_code = 'within'
		elif start_1 > start_2 and end_1 < end_2:
			return_code = 'within'
		elif start_1 <= start_2 and end_1 > end_2:
			return_code = 'contain'
		elif start_1 < start_2 and end_1 >= end_2:
			return_code = 'contain'
		elif start_1 < start_2 and end_1 > end_2:
			return_code = 'contain'
		else:
			return_code = 'overlap'
	return return_code


def srog_peak(inbed1, inbed2, outfile, n_up = 1, n_down = 1, max_dist = 250000000):

	"""
	Calculates SROG code for each region in inbed1

	Parameters
	----------
	inbed1 : str
		Name of a BED file.
	inbed2 : str
		Name of another BED file.
	outfile : str
		Name of output file.

	Returns
	-------
	pd Series
	"""

	maps = {}
	OUTPUT = open(outfile, 'w')
	srog_summary = {'disjoint':0, 'overlap':0, 'contain':0, 'within':0, 'touch':0, 'equal':0, 'other':0}
	logging.info("Build interval tree from file: \"%s\"" % inbed2)
	for l in ireader.reader(inbed2):
		if l.startswith(('browser','#','track')):continue
		f = l.split()
		if len(f) < 3:
			logging.warning("Invalid BED line (Requires at least 3 columns: chrom, start, end): %s" % l)
			continue
		chrom, start, end = f[0], int(f[1]), int(f[2])
		if start > end:
			logging.warning("invalid BED line (start > end): %s" % l)
			continue

		#try to get name from the 4th column
		try:
			name = f[3]
		except:
			name = f[0] + ':' + f[1] + '-' + f[2]

		#try to get strand from the 6th column
		try:
			strandness = f[5]
			if strandness not in ('+','-'):
				strandness = '+'
		except:
			strandness = '+'

		if chrom not in maps:
			maps[chrom] = Intersecter()
		maps[chrom].add_interval( Interval(start, end, value = name, strand = strandness))

	logging.info("Reading BED file: \"%s\"" % inbed1)
	for l in ireader.reader(inbed1):
		if l.startswith(('browser','#','track')):continue
		f = l.split()
		if len(f) < 3:
			logging.warning("Invalid BED line (Requires at least 3 columns: chrom, start, end): %s" % l)
			continue
		chrom, start, end = f[0], int(f[1]), int(f[2])
		if start > end:
			logging.warning("Invalid BED line (start > end): %s" % l)
			continue

		#try to get name from the 4th column
		try:
			name = f[3]
		except:
			name = f[0] + ':' + f[1] + '-' + f[2]

		#try to get strand from the 6th column
		try:
			strandness = f[5]
			if strandness not in ('+','-'):
				strandness = '+'
		except:
			strandness = '+'

		if chrom not in maps:
			srog_summary['disjoint'] += 1
			print (l + '\t' + 'NA' + '\t' + 'NA', file=OUTPUT)
			continue

		overlaps = maps[chrom].find(start, end)
		if len(overlaps) == 0:
			srog_summary['disjoint'] += 1
			up_interval =  maps[chrom].upstream_of_interval(Interval(start, end, strand = strandness), num_intervals = n_up, max_dist = max_dist)
			down_interval =  maps[chrom].downstream_of_interval(Interval(start, end, strand = strandness), num_intervals = n_down, max_dist = max_dist)
			if len(up_interval) == 0:
				up_interval_name = 'NA'
			else:
				up_interval_name = str(up_interval[0].value)
			if len(down_interval) == 0:
				down_interval_name = 'NA'
			else:
				down_interval_name = str(down_interval[0].value)
			print (l + '\t' + 'disjoint' + '\t' + 'UpInterval=' + up_interval_name + ',' + 'DownInterval=' + down_interval_name, file=OUTPUT)
		else:
			srog_codes = []
			target_names = []
			for o in overlaps:
				tmp = srogcode((chrom, start, end), (chrom, o.start, o.end))
				srog_codes.append(tmp)
				target_names.append(o.value)
			print (l + '\t' + ','.join(srog_codes) + '\t' + ','.join(target_names), file=OUTPUT)
			for code in srog_codes:
				srog_summary[code] += 1
	return pd.Series(data=srog_summary)


if __name__=='__main__':
	#(a, b, common) = compare_bed(sys.argv[1], sys.argv[2])
	#bedtofile(a,'a')
	#bedtofile(b,'b')
	#bedtofile(common,'common')
	a = cooccur_peak(sys.argv[1], sys.argv[2], sys.argv[3])
	print (a)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan  5 13:06:22 2022

@author: m102324
"""

import sys
from cobindability.BED import bed_genomic_size, bed_overlap_size, bedtolist, bed_counts
#from cobindability.coefcal import ov_coef, ov_jaccard, ov_ss, ov_sd
from os.path import basename
import logging
import numpy as np
import pandas as pd
from random import sample
from cobindability import version


__author__ = "Liguo Wang"
__copyright__ = "Copyleft"
__credits__ = []
__license__ = "MIT"
__version__ = version.version
__maintainer__ = "Liguo Wang"
__email__ = "wang.liguo@mayo.edu"
__status__ = "Development"


def bootstrap_coef(file1, file2, score_func, size_factor, n_draws=20, fraction = 0.75, bg_size = 1.4e9):
	"""
	Calculate the following indices:
	- overlap coefficient,
	- Jaccard's coefficient (https://en.wikipedia.org/wiki/Jaccard_index),
	- Szymkiewicz–Simpson coefficient, (https://en.wikipedia.org/wiki/Overlap_coefficient),
	- Sørensen–Dice coefficient (https://en.wikipedia.org/wiki/Sørensen–Dice_coefficient),
	- Pointwise mutual information (https://en.wikipedia.org/wiki/Pointwise_mutual_information),

	Parameters
	----------
	file1 : str
		Genomic regions in BED (Browser Extensible Data, https://genome.ucsc.edu/FAQ/FAQformat.html#format1), BED-like or BigBed format.
		The BED-like format includes 'bed3','bed4','bed6','bed12','bedgraph','narrowpeak', 'broadpeak','gappedpeak'. BED and BED-like
		format can be plain text, compressed (.gz, .z, .bz, .bz2, .bzip2) or remote (http://, https://, ftp://) files. Do not compress
		BigBed foramt. BigBed file can also be a remote file.
	file2 : str
		Genomic regions in BED (Browser Extensible Data, https://genome.ucsc.edu/FAQ/FAQformat.html#format1), BED-like or BigBed format.
		The BED-like format includes 'bed3','bed4','bed6','bed12','bedgraph','narrowpeak', 'broadpeak','gappedpeak'. BED and BED-like
		format can be plain text, compressed (.gz, .z, .bz, .bz2, .bzip2) or remote (http://, https://, ftp://) files. Do not compress
		BigBed foramt. BigBed file can also be a remote file.
	score_func : function
		Function to calculate overlap index. Include ov_coef, ov_jaccard, ov_ss, ov_sd, pmi_value, npmi_value
	n_draws : int
		Times of drawing samples with replacement. Set to '0' to turn off bootstraping.  The default is 20.
	fraction : float, optional
		The fraction of subsample. The default is 0.75 (75% of the orignal genomic reginos will be selected).
	bg_size : int, optional
		The effective background genome size. About 1.4Gb of the human genome are. The default is 1.4e9.

	Returns
	-------
	pandas.Series
		Pandas Series of 'coef_obs', 'coef_exp','coef_ratio', 'coef_ratio_low', 'coef_ratio_high'.
		'coef_obs' : Observed overlap coefficient between two BED files.
		'coef_exp' : Expected overlap coefficient between two BED files.
		'coef_ratio' : Ratio between 'coef_obs' and 'coef_exp'.
		'coef_ratio_low' : The lower bound of 95% confidence interval of 'coef_ratio'.
		'coef_ratio_high' : The upper bound of 95% confidence interval of 'coef_ratio'.

	"""

	results = {}

	file1_lst = bedtolist (file1)
	file2_lst = bedtolist (file2)

	results['A.name'] = basename(file1)
	results['B.name'] = basename(file2)

	#calculate interval counts
	logging.debug("Calculating bed counts ...")
	totalCount1, totalCount2 = bed_counts(file1, file2)
	results['A.interval_count'] = totalCount1
	results['B.interval_count'] = totalCount2


	#calculate overall overlap coef
	#logging.info("Calculating coefficient ...")
	(uniqBase1,uniqBase2) = bed_genomic_size(file1_lst, file2_lst)


	overlapBases = bed_overlap_size(file1_lst, file2_lst)
	overlapBases_exp = uniqBase1*uniqBase2/bg_size
	[unionBases] = bed_genomic_size(file1_lst + file2_lst)

	results['A.size'] = uniqBase1
	results['B.size'] = uniqBase2
	results['A_or_B.size'] = unionBases
	results['A_and_B.size'] = overlapBases
	results['Coef'] = score_func(uniqBase1, uniqBase2, overlapBases, bg_size)
	results['Coef(expected)'] = score_func(uniqBase1, uniqBase2, overlapBases_exp, bg_size)

	if n_draws > 0:
		if fraction > 0 and fraction < 1:
			resample_size1 = int(totalCount1 * fraction)
			resample_size2 = int(totalCount2 * fraction)
		else:
			logging.error("Fraction must be > 0 and < 1.")
			sys.exit(0)
		logging.debug("Bootstraping is on. Iterate %d times. " % n_draws)
		tmp = []
		for i in range(n_draws):
			logging.debug("Bootstrap resampling %d ..." % i)
			sample_1 = sample(file1_lst, resample_size1)
			sample_2 = sample(file2_lst, resample_size2)

			sample_overlapBases = bed_overlap_size(sample_1, sample_2)
			if size_factor != 1:
				sample_overlapBases = sample_overlapBases * size_factor
			(sample1_size, sample2_size) = bed_genomic_size(sample_1, sample_2)
			sample_ovcoef = score_func(sample1_size, sample2_size, sample_overlapBases, bg_size)
			tmp.append(sample_ovcoef)
		ci_lower = np.percentile(np.array(tmp), 2.5)
		ci_upper = np.percentile(np.array(tmp), 97.5)
		results['Coef(95% CI)'] = '[%.4f,%.4f]' % (ci_lower, ci_upper)
	else:
		logging.info("Bootstraping is off ...")
		results['Coef(95% CI)'] = '[NA,NA]'

	return pd.Series(data=results)


def bootstrap_npmi(file1, file2, score_func, size_factor, n_draws=20, fraction = 0.75, bg_size = 1.4e9):
	"""
	Calculate the following indices:
	- Normalized pointwise mutual information.

	Parameters
	----------
	file1 : str
		Genomic regions in BED (Browser Extensible Data, https://genome.ucsc.edu/FAQ/FAQformat.html#format1), BED-like or BigBed format.
		The BED-like format includes 'bed3','bed4','bed6','bed12','bedgraph','narrowpeak', 'broadpeak','gappedpeak'. BED and BED-like
		format can be plain text, compressed (.gz, .z, .bz, .bz2, .bzip2) or remote (http://, https://, ftp://) files. Do not compress
		BigBed foramt. BigBed file can also be a remote file.
	file2 : str
		Genomic regions in BED (Browser Extensible Data, https://genome.ucsc.edu/FAQ/FAQformat.html#format1), BED-like or BigBed format.
		The BED-like format includes 'bed3','bed4','bed6','bed12','bedgraph','narrowpeak', 'broadpeak','gappedpeak'. BED and BED-like
		format can be plain text, compressed (.gz, .z, .bz, .bz2, .bzip2) or remote (http://, https://, ftp://) files. Do not compress
		BigBed foramt. BigBed file can also be a remote file.
	score_func : function
		Function to calculate overlap index. Include ov_coef, ov_jaccard, ov_ss, ov_sd, pmi_value, npmi_value
	n_draws : int
		Times of drawing samples with replacement. Set to '0' to turn off bootstraping.  The default is 20.
	fraction : float, optional
		The fraction of subsample. The default is 0.75 (75% of the orignal genomic reginos will be selected).
	bg_size : int, optional
		The effective background genome size. About 1.4Gb of the human genome are. The default is 1.4e9.

	Returns
	-------
	pandas.Series
		Pandas Series of 'coef_obs', 'coef_exp','coef_ratio', 'coef_ratio_low', 'coef_ratio_high'.
		'coef_obs' : Observed overlap coefficient between two BED files.
		'coef_exp' : Expected overlap coefficient between two BED files.
		'coef_ratio' : Ratio between 'coef_obs' and 'coef_exp'.
		'coef_ratio_low' : The lower bound of 95% confidence interval of 'coef_ratio'.
		'coef_ratio_high' : The upper bound of 95% confidence interval of 'coef_ratio'.

	"""

	results = {}

	file1_lst = bedtolist (file1)
	file2_lst = bedtolist (file2)

	results['A.name'] = basename(file1)
	results['B.name'] = basename(file2)

	#calculate interval counts
	logging.debug("Calculating bed counts ...")
	totalCount1, totalCount2 = bed_counts(file1, file2)
	results['A.interval_count'] = totalCount1
	results['B.interval_count'] = totalCount2


	#calculate overall overlap coef
	#logging.info("Calculating coefficient ...")
	(uniqBase1,uniqBase2) = bed_genomic_size(file1_lst, file2_lst)


	overlapBases = bed_overlap_size(file1_lst, file2_lst)
	overlapBases_exp = uniqBase1*uniqBase2/bg_size
	[unionBases] = bed_genomic_size(file1_lst + file2_lst)

	results['A.size'] = uniqBase1
	results['B.size'] = uniqBase2
	results['A_or_B.size'] = unionBases
	results['A_and_B.size'] = overlapBases
	results['Coef'] = score_func(uniqBase1, uniqBase2, overlapBases, bg_size)
	results['Coef(expected)'] = score_func(uniqBase1, uniqBase2, overlapBases_exp, bg_size)

	if n_draws > 0:
		if fraction > 0 and fraction < 1:
			resample_size1 = int(totalCount1 * fraction)
			resample_size2 = int(totalCount2 * fraction)
		else:
			logging.error("Fraction must be > 0 and < 1.")
			sys.exit(0)
		logging.debug("Bootstraping is on. Iterate %d times. " % n_draws)
		tmp = []
		for i in range(n_draws):
			logging.debug("Bootstrap resampling %d ..." % i)
			sample_1 = sample(file1_lst, resample_size1)
			sample_2 = sample(file2_lst, resample_size2)

			sample_overlapBases = bed_overlap_size(sample_1, sample_2)
			if size_factor != 1:
				sample_overlapBases = sample_overlapBases * size_factor
			(sample1_size, sample2_size) = bed_genomic_size(sample_1, sample_2)
			sample_ovcoef = score_func(sample1_size, sample2_size, sample_overlapBases, bg_size*fraction)
			tmp.append(sample_ovcoef)
		ci_lower = np.percentile(np.array(tmp), 2.5)
		ci_upper = np.percentile(np.array(tmp), 97.5)
		results['Coef(95% CI)'] = '[%.4f,%.4f]' % (ci_lower, ci_upper)
	else:
		logging.info("Bootstraping is off ...")
		results['Coef(95% CI)'] = '[NA,NA]'

	return pd.Series(data=results)



#!/usr/bin/env python
# encoding: utf-8

"""
Read regular, compressed (.gz .bz), remote (http://, https://, ftp://) BED file or BigBed file.
"""


import sys
import bz2
import gzip
from urllib.request import urlopen
from subprocess import Popen, PIPE
import pyBigWig
from cobindability import version

__author__ = "Liguo Wang"
__copyright__ = "Copyleft"
__credits__ = []
__license__ = "MIT"
__version__ = version.version
__maintainer__ = "Liguo Wang"
__email__ = "wang.liguo@mayo.edu"
__status__ = "Development"



def bbopen(fname):
	"""
	Open bigBed file. Local and remote bigBed read access is supported.
	"""
	bb = pyBigWig.open(fname)
	chrom_dict = bb.chroms()
	for chr in chrom_dict:
		for start, end, score in bb.entries(chr, 0, chrom_dict[chr]):
			yield(chr + '\t' + str(start) + '\t' + str(end) + '\t' + score)


def nopen(f, mode="rb"):
	"""
	Open regular or compressed BED file.
	"""
	if not isinstance(f, str):
		return f
	if f.startswith("|"):
		p = Popen(f[1:], stdout=PIPE, stdin=PIPE, shell=True)
		if mode[0] == "r": return p.stdout
		return p
	return {"r": sys.stdin, "w": sys.stdout}[mode[0]] if f == "-" \
		else gzip.open(f, mode) if f.endswith((".gz", ".Z", ".z")) \
		else bz2.BZ2File(f, mode) if f.endswith((".bz", ".bz2", ".bzip2")) \
		else urlopen(f) if f.startswith(("http://", "https://","ftp://")) \
		else open(f, mode)


def reader(fname):
	if fname.endswith(('.bb','.bigbed','.bigBed','.BigBed', '.BB',' BIGBED')):
		for l in bbopen(fname):
			yield l
	else:
		for l in nopen(fname):
			yield l.decode('utf8').strip().replace("\r", "")



import sys, os, platform, glob
from distutils.core import setup
from setuptools import *

"""
Setup script for cobind  -- collocation analysis of genomics regions.
"""


def main():
    setup(  name = "cobind",
            version = "1.0.0",
            python_requires='>=3.5',
            py_modules = [ 'psyco_full' ],
            packages = find_packages( 'lib' ),
            package_dir = { '': 'lib' },
            package_data = { '': ['*.ps'] },
            scripts = glob.glob( "bin/*.py"),
            ext_modules = [],
            test_suite = 'nose.collector',
            setup_requires = ['nose>=0.10.4','cython>=0.12'],
            author = "Liguo Wang",
            author_email ="wangliguo78@gmail.com",
            platforms = ['Linux','MacOS'],
            requires = [],
            install_requires = ['scipy', 'numpy', 'pandas', 'bx-python','pyBigWig'],
            description = "collocation analysis of genomics regions",
            url = "https://cobind.readthedocs.io/en/latest/",
            zip_safe = False,
            dependency_links = [],
            classifiers=[
                'Development Status :: 4 - Beta',
                'Environment :: Console',
                'Intended Audience :: Science/Research',
                'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
                'Operating System :: MacOS :: MacOS X',
                'Operating System :: POSIX',
                'Programming Language :: Python',
                'Topic :: Scientific/Engineering :: Bio-Informatics',
            ],

            keywords='genome regions, overlap coefficient, cooccur, pointwise mutual information',
             )


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import sys
import logging
import argparse
import pandas as pd
from cobindability.BED import compare_bed,peakwise_ovcoef,cooccur_peak,srog_peak
#from cobindability.ovpmi import cal_pmi
from cobindability.bw import bigwig_corr
from cobindability.ovstat import ov_stats
from cobindability import version
from cobindability.ovbootstrap import bootstrap_coef,bootstrap_npmi
from cobindability.coefcal import ov_coef, ov_jaccard, ov_ss, ov_sd, pmi_value, npmi_value




__author__ = "Liguo Wang"
__copyright__ = "Copyleft"
__credits__ = []
__license__ = "GPL"
__version__ = version.version
__maintainer__ = "Liguo Wang"
__email__ = "wang.liguo@mayo.edu"
__status__ = "Development"

def main():
	pd.set_option('display.float_format', lambda x: '%.4f' % x)
	general_help = "**cobind: collocation analyses of genomic regions**"
	bed_help = "Genomic regions in BED, BED-like or bigBed format. The BED-like format includes:\
		'bed3', 'bed4', 'bed6', 'bed12', 'bedgraph', 'narrowpeak', 'broadpeak', 'gappedpeak'.\
		BED and BED-like format can be plain text, compressed (.gz, .z, .bz, .bz2, .bzip2) \
		or remote (http://, https://, ftp://) files. Do not compress BigBed foramt.\
		BigBed file can also be a remote file."
	# sub commands and help.
	commands = {
	'overlap' : "Calculate the overlapping coefficient (O) between two sets of genomic regions. O = |A and B| / (|A|*|B|)**0.5",
	'jaccard' : "Calculate the Jaccard similarity coefficient (J) between two sets of genomic regions. J = |A and B| / |A or B|",
	'dice' : "Calculate the Sørensen–Dice coefficient (SD) between two sets of genomic regions. SD = 2*|A and B| / (|A| + |B|)",
	'simpson' : "Calculate the Szymkiewicz–Simpson coefficient (SS) between two sets of genomic regions. SS = |A and B| / min(|A|, |B|)",
	'pmi' : "Calculate the pointwise mutual information (PMI) between two sets of genomic regions. PMI = log(p(|A and B|)) - log(p(|A|)) - log(p(|B|))",
	'npmi' : "Calculate the normalized pointwise mutual information (NPMI) between two sets of genomic regions. NPMI = log(p(|A|)*p(|B|)) / log(p(|A and B|)) - 1",
	'cooccur' : "Evaluate if two sets of genomic regions are significantly overlapped in given background regions.",
	'covary' : "Calculate the covariance (Pearson, Spearman and Kendall coefficients) of binding intensities between two sets of genomic regions.",
	'srog' : "Report the code of Spatial Relation Of Genomic (SROG) regions. SROG codes include 'disjoint','touch','equal','overlap','contain','within'.",
	'stat' : "Wrapper function. Report basic statistics of genomic regions, and calculate overlapping measurements, including \"O\", \"J\", \"SD\", \"SS\", \"PMI\", \"NPMI\", without bootstrap resampling or generating peakwise measurements.",
	}

	#create parse
	parser = argparse.ArgumentParser(description=general_help, epilog='', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
	parser.add_argument('-v', '--version', action='version', version='%s %s' %  ('cobind', __version__))

	# create sub-parser
	sub_parsers = parser.add_subparsers(help='Sub-command description:')
	parser_overlap = sub_parsers.add_parser('overlap', help=commands['overlap'])
	parser_jaccard = sub_parsers.add_parser('jaccard', help=commands['jaccard'])
	parser_dice = sub_parsers.add_parser('dice', help=commands['dice'])
	parser_simpson = sub_parsers.add_parser('simpson', help=commands['simpson'])
	parser_pmi = sub_parsers.add_parser('pmi', help=commands['pmi'])
	parser_npmi = sub_parsers.add_parser('npmi', help=commands['npmi'])
	parser_cooccur = sub_parsers.add_parser('cooccur', help=commands['cooccur'])
	parser_covary = sub_parsers.add_parser('covary', help=commands['covary'])
	parser_srog = sub_parsers.add_parser('srog', help=commands['srog'])
	parser_stat = sub_parsers.add_parser('stat', help=commands['stat'])


	# create the parser for the "overlap" sub-command
	parser_overlap.add_argument("bed1", type=str, metavar ="input_A.bed",help=bed_help)
	parser_overlap.add_argument("bed2", type=str, metavar ="input_B.bed",help=bed_help)
	parser_overlap.add_argument('-n', '--ndraws', type=int, dest="iter", default = 20, help="Times of resampling to estimate confidence intervals. Set to '0' to turn off resampling.(default: %(default)d)")
	parser_overlap.add_argument('-f', '--fraction', type=int, dest="subsample", default = 0.75, help="Resampling fraction. (default: %(default).2f)")
	parser_overlap.add_argument('-b', '--background', type=int, dest="bgsize", default = 1.4e9, help="The size of the cis-regulatory genomic regions. This is about 1.4Gb For the human genome. (default: %(default)d)")
	parser_overlap.add_argument("-o", "--save", action="store_true", help="If set, will save peak-wise coefficients to files (\"input_A_peakwise_scores.tsv\" and \"input_B_peakwise_scores.tsv\").")
	parser_overlap.add_argument("-d", "--debug",action="store_true", help="Print detailed information for debugging.")


	# create the parser for the "jaccard" sub-command
	parser_jaccard.add_argument("bed1", type=str, metavar ="input_A.bed",help=bed_help)
	parser_jaccard.add_argument("bed2", type=str, metavar ="input_B.bed",help=bed_help)
	parser_jaccard.add_argument('-n', '--ndraws', type=int, dest="iter", default = 20, help="Times of resampling to estimate confidence intervals. Set to '0' to turn off resampling.(default: %(default)d)")
	parser_jaccard.add_argument('-f', '--fraction', type=int, dest="subsample", default = 0.75, help="Resampling fraction. (default: %(default).2f)")
	parser_jaccard.add_argument('-b', '--background', type=int, dest="bgsize", default = 1.4e9, help="The size of the cis-regulatory genomic regions. This is about 1.4Gb For the human genome. (default: %(default)d)")
	parser_jaccard.add_argument("-o", "--save", action="store_true", help="If set, will save peak-wise coefficients to files (\"input_A_peakwise_scores.tsv\" and \"input_B_peakwise_scores.tsv\").")
	parser_jaccard.add_argument("-d", "--debug",action="store_true", help="Print detailed information for debugging.")

	# create the parser for the "dice" sub-command
	parser_dice.add_argument("bed1", type=str, metavar ="input_A.bed",help=bed_help)
	parser_dice.add_argument("bed2", type=str, metavar ="input_B.bed",help=bed_help)
	parser_dice.add_argument('-n', '--ndraws', type=int, dest="iter", default = 20, help="Times of resampling to estimate confidence intervals. Set to '0' to turn off resampling.(default: %(default)d)")
	parser_dice.add_argument('-f', '--fraction', type=int, dest="subsample", default = 0.75, help="Resampling fraction. (default: %(default).2f)")
	parser_dice.add_argument('-b', '--background', type=int, dest="bgsize", default = 1.4e9, help="The size of the cis-regulatory genomic regions. This is about 1.4Gb For the human genome. (default: %(default)d)")
	parser_dice.add_argument("-o", "--save", action="store_true", help="If set, will save peak-wise coefficients to files (\"input_A_peakwise_scores.tsv\" and \"input_B_peakwise_scores.tsv\").")
	parser_dice.add_argument("-d", "--debug",action="store_true", help="Print detailed information for debugging.")


	# create the parser for the "simpson" sub-command
	parser_simpson.add_argument("bed1", type=str, metavar ="input_A.bed",help=bed_help)
	parser_simpson.add_argument("bed2", type=str, metavar ="input_B.bed",help=bed_help)
	parser_simpson.add_argument('-n', '--ndraws', type=int, dest="iter", default = 20, help="Times of resampling to estimate confidence intervals. Set to '0' to turn off resampling.(default: %(default)d)")
	parser_simpson.add_argument('-f', '--fraction', type=int, dest="subsample", default = 0.75, help="Resampling fraction. (default: %(default).2f)")
	parser_simpson.add_argument('-b', '--background', type=int, dest="bgsize", default = 1.4e9, help="The size of the cis-regulatory genomic regions. This is about 1.4Gb For the human genome. (default: %(default)d)")
	parser_simpson.add_argument("-o", "--save", action="store_true", help="If set, will save peak-wise coefficients to files (\"input_A_peakwise_scores.tsv\" and \"input_B_peakwise_scores.tsv\").")
	parser_simpson.add_argument("-d", "--debug",action="store_true", help="Print detailed information for debugging.")

	# create the parser for the "pmi" sub-command
	parser_pmi.add_argument("bed1", type=str, metavar ="input_A.bed",help=bed_help)
	parser_pmi.add_argument("bed2", type=str, metavar ="input_B.bed",help=bed_help)
	parser_pmi.add_argument('-n', '--ndraws', type=int, dest="iter", default = 20, help="Times of resampling to estimate confidence intervals. Set to '0' to turn off resampling.(default: %(default)d)")
	parser_pmi.add_argument('-f', '--fraction', type=int, dest="subsample", default = 0.75, help="Resampling fraction. (default: %(default).2f)")
	parser_pmi.add_argument('-b', '--background', type=int, dest="bgsize", default = 1.4e9, help="The size of the cis-regulatory genomic regions. This is about 1.4Gb For the human genome. (default: %(default)d)")
	parser_pmi.add_argument("-o", "--save", action="store_true", help="If set, will save peak-wise coefficients to files (\"input_A_peakwise_scores.tsv\" and \"input_B_peakwise_scores.tsv\").")
	parser_pmi.add_argument("-d", "--debug",action="store_true", help="Print detailed information for debugging.")

	# create the parser for the "npmi" sub-command
	parser_npmi.add_argument("bed1", type=str, metavar ="input_A.bed",help=bed_help)
	parser_npmi.add_argument("bed2", type=str, metavar ="input_B.bed",help=bed_help)
	parser_npmi.add_argument('-n', '--ndraws', type=int, dest="iter", default = 20, help="Times of resampling to estimate confidence intervals. Set to '0' to turn off resampling.(default: %(default)d)")
	parser_npmi.add_argument('-f', '--fraction', type=int, dest="subsample", default = 0.75, help="Resampling fraction. (default: %(default).2f)")
	parser_npmi.add_argument('-b', '--background', type=int, dest="bgsize", default = 1.4e9, help="The size of the cis-regulatory genomic regions. This is about 1.4Gb For the human genome. (default: %(default)d)")
	parser_npmi.add_argument("-o", "--save", action="store_true", help="If set, will save peak-wise coefficients to files (\"input_A_peakwise_scores.tsv\" and \"input_B_peakwise_scores.tsv\").")
	parser_npmi.add_argument("-d", "--debug",action="store_true", help="Print detailed information for debugging.")


	# create the parser for the "cooccur" sub-command
	parser_cooccur.add_argument("bed1", type=str, metavar ="input_A.bed",help=bed_help)
	parser_cooccur.add_argument("bed2", type=str, metavar ="input_B.bed",help=bed_help)
	parser_cooccur.add_argument("bed3", type=str, metavar ="background.bed",help="Genomic regions as the background (e.g., all promoters, all enhancers).")
	parser_cooccur.add_argument("output", type=str, metavar ="output.tsv",help="For each genomic region in the \"background.bed\" file, add another column indicating if this region is \"input_A specific (i.e., A+B-)\", \"input_B specific (i.e., A-B+)\", \"co-occur (i.e., A+B+)\" or \"neither (i.e, A-B-)\". ")
	parser_cooccur.add_argument('--ncut', type=int, dest="n_cut",  default = 1, help="The minimum overlap size. (default: %(default)d)")
	parser_cooccur.add_argument('--pcut', type=float, dest="p_cut",  default = 0.0, help="The minimum overlap percentage. (default: %(default)f)")
	parser_cooccur.add_argument("-d", "--debug",action="store_true", help="Print detailed information for debugging.")

	# create the parser for the "covary" sub-command
	parser_covary.add_argument("bed1", type=str, metavar="input_A.bed",help=bed_help)
	parser_covary.add_argument("bw1", type=str, metavar="input_A.bw",help="Input bigWig file matched to 'input_A.bed'. BigWig file can be local or remote. Note: the chromosome IDs must be consistent between BED and bigWig files.")
	parser_covary.add_argument("bed2", type=str, metavar="input_B.bed",help=bed_help)
	parser_covary.add_argument("bw2", type=str, metavar="input_B.bw",help="Input bigWig file matched to 'input_B.bed'. BigWig file can be local or remote.  Note: the chromosome IDs must be consistent between BED and bigWig files.")
	parser_covary.add_argument("output", type=str,metavar="output_prefix", help="Prefix of output files. Three files will be generated: \"output_prefix_bedA_unique.tsv\" (input_A.bed specific regions and their bigWig scores), \"output_prefix_bedB_unique.tsv\" (input_B.bed specific regions and their bigWig scores), and \"output_prefix_common.tsv\"(input_A.bed and input_B.bed overlapped regions and their bigWig scores).")
	parser_covary.add_argument("--na", type=str, dest="na_label", default = 'nan', help="Symbols used to represent the missing values. (default: %(default)s)")
	parser_covary.add_argument('--type', type=str, dest="score_type", choices=['mean', 'min', 'max'],  default = 'mean', help="Summary statistic score type ('min','mean' or 'max') of a genomic region. (default: %(default)s)")
	parser_covary.add_argument('--topx', type=float, dest="top_X",  default = 1.0, help="Fraction (if 0 < top_X <= 1) or number (if top_X > 1) of genomic regions used to calculate Pearson, Spearman, Kendall's correlations. If TOP_X == 1 (i.e., 100%%), all the genomic regions will be used to calculate correlations. (default: %(default)s)")
	parser_covary.add_argument('--min_sig', type=float, dest="min_signal", default = 0, help="Genomic region with summary statistic score <= this will be removed. (default: %(default)s)")
	parser_covary.add_argument("--exact", dest="exact", action="store_true", help="If set, calculate the \"exact\" summary statistic score rather than \"zoom-level\" score for each genomic region.")
	parser_covary.add_argument("--keepna", dest="keepna", action="store_true", help="If set, a genomic region will be kept even it does not have summary statistical score in either of the two bigWig files. This flag only affects the output TSV files.")
	parser_covary.add_argument("-d", "--debug",action="store_true", help="Print detailed information for debugging.")

	# create the parser for the "srog" sub-command
	parser_srog.add_argument("bed1", type=str, metavar ="input_A.bed",help="Genomic regions in BED, BED-like or bigBed format. If 'name' (the 4th column) is not provided, the default name is \"chrom:start-end\". If strand (the 6th column) is not provided, the default strand is \"+\".")
	parser_srog.add_argument("bed2", type=str, metavar ="input_B.bed",help="Genomic regions in BED, BED-like or bigBed format. If 'name' (the 4th column) is not provided, the default name is \"chrom:start-end\". If strand (the 6th column) is not provided, the default strand is \"+\". ")
	parser_srog.add_argument("output", type=str, metavar ="output.tsv",help="Generate spatial relation code (disjoint, touch, equal, overlap, contain, within) for each genomic interval in \"input_A.bed\".")
	parser_srog.add_argument('--dist', type=int, dest="max_dist",  default = 250000000, help="When intervals are disjoint, find the closest up- and down-stream intervals that are no further than `max_dist` away. default: %(default)d)")
	parser_srog.add_argument("-d", "--debug",action="store_true", help="Print detailed information for debugging.")


	# create the parser for the "stat" sub-command
	parser_stat.add_argument("bed1", type=str, metavar ="input_A.bed",help=bed_help)
	parser_stat.add_argument("bed2", type=str, metavar ="input_B.bed",help=bed_help)
	parser_stat.add_argument('-b', '--background', type=int, dest="bgsize", default = 1.4e9, help="The size of the cis-regulatory genomic regions. This is about 1.4Gb For the human genome. (default: %(default)d)")
	parser_stat.add_argument("-d", "--debug",action="store_true", help="Print detailed information for debugging.")


	args = parser.parse_args()


	if len(sys.argv)==1:
		parser.print_help(sys.stderr)
		sys.exit(0)
	elif len(sys.argv) >= 2:
		command = sys.argv[1]
		if command =='stat':
			if args.debug:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.DEBUG)
			else:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.INFO)

			info = ov_stats(args.bed1, args.bed2, bg_size = args.bgsize)
			print (info)

		elif command == 'overlap':
			if args.debug:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.DEBUG)
			else:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.INFO)
			logging.info("Calculate overlapping coefficient (overall) ...")
			result = bootstrap_coef(args.bed1, args.bed2, size_factor = 1/args.subsample, score_func = ov_coef, n_draws = args.iter, fraction = args.subsample, bg_size = args.bgsize)
			print (result)
			if args.save:
				logging.info("Calculate overlapping coefficient (peak-wise) ...")
				peakwise_ovcoef(args.bed1, args.bed2, score_func = ov_coef, g = args.bgsize, na_label='NA')

		elif command == 'jaccard':
			if args.debug:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.DEBUG)
			else:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.INFO)
			logging.info("Calculate Jaccard coefficient (overall) ...")
			result = bootstrap_coef(args.bed1, args.bed2, size_factor = 1/args.subsample, score_func = ov_jaccard, n_draws = args.iter, fraction = args.subsample, bg_size = args.bgsize)
			print (result)
			if args.save:
				logging.info("Calculate Jaccard coefficient (peakwise) ...")
				peakwise_ovcoef(args.bed1, args.bed2, score_func = ov_jaccard, g = args.bgsize, na_label='NA')

		elif command == 'dice':
			if args.debug:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.DEBUG)
			else:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.INFO)
			logging.info("Calculate Sørensen–Dice coefficient (overall) ...")
			result = bootstrap_coef(args.bed1, args.bed2, size_factor = 1/args.subsample, score_func = ov_sd, n_draws = args.iter, fraction = args.subsample, bg_size = args.bgsize)
			print (result)
			if args.save:
				logging.info("Calculate Sørensen–Dice coefficient (peakwise) ...")
				peakwise_ovcoef(args.bed1, args.bed2, score_func = ov_sd, g = args.bgsize, na_label = 'NA')

		elif command == 'simpson':
			if args.debug:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.DEBUG)
			else:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.INFO)
			logging.info("Calculate Szymkiewicz–Simpson coefficient (overall) ...")
			result = bootstrap_coef(args.bed1, args.bed2, size_factor = 1/args.subsample, score_func = ov_ss, n_draws = args.iter, fraction = args.subsample, bg_size = args.bgsize)
			print (result)
			if args.save:
				logging.info("Calculate Szymkiewicz–Simpson coefficient (peakwise) ...")
				peakwise_ovcoef(args.bed1, args.bed2, score_func = ov_ss, g = args.bgsize, na_label = 'NA')

		elif command == 'pmi':
			if args.debug:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.DEBUG)
			else:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.INFO)
			logging.info("Calculate the pointwise mutual information (PMI) ...")
			#result = cal_pmi(args.bed1, args.bed2, bg_size = args.bgsize)
			result = bootstrap_coef(args.bed1, args.bed2, size_factor = 1, score_func = pmi_value, n_draws = args.iter, fraction = args.subsample, bg_size = args.bgsize)
			print (result)
			if args.save:
				peakwise_ovcoef(args.bed1, args.bed2, score_func = pmi_value, g = args.bgsize, na_label = 'NA')

		elif command == 'npmi':
			if args.debug:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.DEBUG)
			else:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.INFO)
			logging.info("Calculate the normalized pointwise mutual information (NPMI) ...")
			#result = cal_pmi(args.bed1, args.bed2, bg_size = args.bgsize)
			result = bootstrap_npmi(args.bed1, args.bed2, size_factor = 1/args.subsample, score_func = npmi_value, n_draws = args.iter, fraction = args.subsample, bg_size = args.bgsize)
			print (result)
			if args.save:
				peakwise_ovcoef(args.bed1, args.bed2, score_func = npmi_value, g = args.bgsize, na_label = 'NA')

		elif command == 'srog':
			if args.debug:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.DEBUG)
			else:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.INFO)
			logging.info("Determine the spacial realtions of genomic (SROG) intervals ...")
			summary = srog_peak(inbed1 = args.bed1, inbed2 = args.bed2, outfile = args.output, max_dist =  args.max_dist)
			print (summary)

		elif command == 'covary':
			if args.debug:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.DEBUG)
			else:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.INFO)
			a_uniq_lst,b_uniq_lst,common_lst = compare_bed(args.bed1, args.bed2)
			logging.info("Calculate covariabilities of overlapped regions ...")
			c_corr = bigwig_corr(bed = common_lst, bw1 = args.bw1, bw2 = args.bw2, outfile = args.output + '_common.tsv', na_label = args.na_label, score_type = args.score_type, exact_scores = args.exact, keep_NA = args.keepna, top_x = args.top_X, min_sig = args.min_signal)
			print (c_corr.T)

			logging.info("Calculate covariabilities of \"%s\" unique regions ..." % args.bed1)
			a_corr = bigwig_corr(bed = a_uniq_lst, bw1 = args.bw1, bw2 = args.bw2, outfile = args.output + '_bedA_unique.tsv', na_label = args.na_label, score_type = args.score_type, exact_scores = args.exact,  keep_NA = args.keepna, top_x = args.top_X, min_sig = args.min_signal)
			print (a_corr.T)

			logging.info("Calculate covariabilities of \"%s\" unique regions ..." % args.bed2)
			b_corr = bigwig_corr(bed = b_uniq_lst, bw1 = args.bw1, bw2 = args.bw2, outfile = args.output + '_bedB_unique.tsv', na_label = args.na_label, score_type = args.score_type, exact_scores = args.exact,  keep_NA = args.keepna, top_x = args.top_X, min_sig = args.min_signal)
			print (b_corr.T)

		elif command == 'cooccur':
			if args.debug:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.DEBUG)
			else:
				logging.basicConfig(format = "%(asctime)s [%(levelname)s]  %(message)s",datefmt='%Y-%m-%d %I:%M:%S', level=logging.INFO)
			logging.info("Calculate the co-occurrence of two sets of genomic intervals ...")
			results = cooccur_peak(inbed1 = args.bed1, inbed2 = args.bed2, inbed_bg = args.bed3, outfile = args.output, n_cut = args.n_cut, p_cut = args.p_cut)
			print (results)


if __name__ == '__main__':
	main()
